--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2007 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <stefan.fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH THE CESYS EFM01
-- USB CARD (ARTICLE-NR.: C1050-4107) ONLY! YOU ARE NOT ALLOWED TO MODIFY AND
-- DISTRIBUTE OR USE IT WITH ANY OTHER HARDWARE, SOFTWARE OR ANY OTHER KIND OF
-- ASIC OR PROGRAMMABLE LOGIC DESIGN WITHOUT THE EXPLICIT PERMISSION OF THE
-- COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;


entity efm01_perf is
  generic
  (
    loopback_mode : boolean := false;
    nr_of_dbgports : positive := 1
  );
  port
  (
    pin_sys_reset_n_i : in std_logic;
    
    pin_fx2_ifclk_i : in std_logic;
    pin_fx2_flag_n_i : in std_logic_vector(2 downto 0);
    pin_fx2_sloe_n_o : out std_logic;
    pin_fx2_slrd_n_o : out std_logic;
    pin_fx2_slwr_n_o : out std_logic;
    pin_fx2_pktend_n_o : out std_logic;
    pin_fx2_fifoadr_o : out std_logic_vector(1 downto 0);
    pin_fx2_fd_io : inout std_logic_vector(15 downto 0);

    pin_debug_io : inout std_logic_vector((nr_of_dbgports-1) downto 0)
  );
end efm01_perf;


architecture RTL of efm01_perf is

  signal sys_reset_n : std_logic := '0';
  
  signal fx2_ifclk : std_logic := '0';
  signal fx2_flag_n : std_logic_vector(2 downto 0) := (others => '1');
  signal fx2_sloe_n : std_logic := '1';
  signal fx2_slrd_n : std_logic := '1';
  signal fx2_slwr_n : std_logic := '1';
  signal fx2_pktend_n : std_logic := '1';
  signal fx2_fifoadr : std_logic_vector(1 downto 0) := (others => '0');

  component fx2_slfifo_ctrl is
    generic
    (
      use_xil_dram : boolean := false;
      nr_of_dbgports : positive := 1
    );
    port
    (
      rst_i : in std_logic;
    
      -- ports connected to cypress fx2 device
      fx2_ifclk_i : in std_logic;
      fx2_flag_n_i : in std_logic_vector(2 downto 0);
      fx2_sloe_n_o : out std_logic;
      fx2_slrd_n_o : out std_logic;
      fx2_slwr_n_o : out std_logic;
      fx2_pktend_n_o : out std_logic;
      fx2_fifoadr_o : out std_logic_vector(1 downto 0);
      fx2_fd_io : inout std_logic_vector(15 downto 0);

      -- ports connected to user application
      app_fifo_wr_i : in std_logic;
      app_fifo_wr_data_i : in std_logic_vector(15 downto 0);
      app_fifo_wr_full_o : out std_logic;
      app_fifo_wr_count_o : out std_logic_vector(3 downto 0);
      app_usb_p2h_pktcount_o : out std_logic_vector(7 downto 0);

      app_fifo_rd_i : in std_logic;
      app_fifo_rd_data_o : out std_logic_vector(15 downto 0);
      app_fifo_rd_empty_o : out std_logic;
      app_fifo_rd_count_o : out std_logic_vector(3 downto 0);
      app_usb_h2p_pktcount_o : out std_logic_vector(7 downto 0);
          
      debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
    );
  end component;
    
  signal rst : std_logic := '1';
  signal clk : std_logic := '0';

  signal app_fifo_wr : std_logic := '0';
  signal app_fifo_wr_data : std_logic_vector(15 downto 0) := (others => '0');
  signal app_fifo_wr_full : std_logic := '0';
  signal app_fifo_wr_count : std_logic_vector(3 downto 0) := (others => '0');
  signal app_usb_p2h_pktcount : std_logic_vector(7 downto 0) := (others => '0');

  signal app_fifo_rd : std_logic := '0';
  signal app_fifo_rd_data : std_logic_vector(15 downto 0) := (others => '0');
  signal app_fifo_rd_empty : std_logic := '0';
  signal app_fifo_rd_count : std_logic_vector(3 downto 0) := (others => '0');
  signal app_usb_h2p_pktcount : std_logic_vector(7 downto 0) := (others => '0');

  signal counter : unsigned(15 downto 0) := (others => '0');

begin

  -- connect entity ports to internal signals
  -------------------------------------------

  sys_reset_n <= pin_sys_reset_n_i;
  
  fx2_ifclk <= pin_fx2_ifclk_i;
  fx2_flag_n <= pin_fx2_flag_n_i;
  pin_fx2_sloe_n_o <= fx2_sloe_n;
  pin_fx2_slrd_n_o <= fx2_slrd_n;
  pin_fx2_slwr_n_o <= fx2_slwr_n;
  pin_fx2_pktend_n_o <= fx2_pktend_n;
  pin_fx2_fifoadr_o <= fx2_fifoadr; 
  
  pin_debug_io <=
  (
    others => 'Z'
  );

  -- instantiate low level slave fifo interface controller
  --------------------------------------------------------

  inst_fx2_slfifo_ctrl : fx2_slfifo_ctrl
  generic map
  (
    use_xil_dram => true,
    nr_of_dbgports => 1
  )
  port map
  (
    rst_i => rst,

    -- ports connected to cypress fx2 device
    fx2_ifclk_i => fx2_ifclk,
    fx2_flag_n_i => fx2_flag_n,
    fx2_sloe_n_o => fx2_sloe_n,
    fx2_slrd_n_o => fx2_slrd_n,
    fx2_slwr_n_o => fx2_slwr_n,
    fx2_pktend_n_o => fx2_pktend_n,
    fx2_fifoadr_o => fx2_fifoadr,
    fx2_fd_io => pin_fx2_fd_io,

    -- ports connected to user application   
    app_fifo_wr_i => app_fifo_wr,
    app_fifo_wr_data_i => app_fifo_wr_data,
    app_fifo_wr_full_o => app_fifo_wr_full,
    app_fifo_wr_count_o => app_fifo_wr_count,
    app_usb_p2h_pktcount_o => app_usb_p2h_pktcount,

    app_fifo_rd_i => app_fifo_rd,
    app_fifo_rd_data_o => app_fifo_rd_data,
    app_fifo_rd_empty_o => app_fifo_rd_empty,
    app_fifo_rd_count_o => app_fifo_rd_count,
    app_usb_h2p_pktcount_o => app_usb_h2p_pktcount, 
    
    debug => open
  );
  
  -- architecture implementation
  ------------------------------
  
  rst <= not sys_reset_n;
  clk <= fx2_ifclk;

  process
  (
    rst, clk,
    app_fifo_rd_data, app_fifo_wr_count, app_fifo_rd_empty, counter, app_fifo_wr_full
  )
  begin
  
    -- application fifo interfaces are simply connected, if synthesized in data
    -- loopback mode
    -- do not to transceive more than 4 kbyte from/to host side in this mode at once,
    -- because fpga buffers inside fx2_slfifo_ctrl are negligible small (30 bytes
    -- each direction) and buffers inside cypress fx2 usb peripheral controller
    -- are 2 kbyte deep (each direction)
    -- considering one clk cycle delay between read strobe and read data valid
    -- in fpga fifo interfaces
    -- => write strobe one clk cycle delay to read strobe
    --    almost full (fifo count < 14) used instead of full flag
    if loopback_mode then
      if rst = '1' then
        app_fifo_wr <= '0';
      elsif clk'event and clk = '1' then
        app_fifo_wr <= app_fifo_rd;
      end if;
      app_fifo_wr_data <= app_fifo_rd_data;
      if (unsigned(app_fifo_wr_count) < 14) and (app_fifo_rd_empty = '0') then
        app_fifo_rd <= '1';
      else
        app_fifo_rd <= '0';
      end if;
      
    -- fpga acts as an endless data source/sink, if synthesized in the other
    -- mode
    -- data is read out (and ignored) at once, if available (host => peripheral dir)
    -- starting at zero, 16 bit counting data is written as soon/long buffer space
    -- is available (peripheral => host dir)
    else
      if rst = '1' then
        counter <= (others => '0');
      elsif clk'event and clk = '1' then
        if app_fifo_wr = '1' then
          counter <= counter + 1;
        end if;
      end if;
      app_fifo_wr_data <= std_logic_vector(counter);
      app_fifo_wr <= not app_fifo_wr_full;
      app_fifo_rd <= not app_fifo_rd_empty;
    end if;
    
  end process;

end RTL;