--------------------------------------------------------------------------------
-- copyright notice:
--------------------------------------------------------------------------------
-- copyright (c) 2007 by
--
-- CESYS GmbH / GERMANY
--
-- http://www.cesys.com
--------------------------------------------------------------------------------
-- author:
--------------------------------------------------------------------------------
-- Stefan Fischer <stefan.fischer@cesys.com>
--------------------------------------------------------------------------------
-- license:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS NOT FREE! IT IS FOR USE TOGETHER WITH THE CESYS EFM01
-- USB CARD (ARTICLE-NR.: C1050-4107) ONLY! YOU ARE NOT ALLOWED TO MODIFY AND
-- DISTRIBUTE OR USE IT WITH ANY OTHER HARDWARE, SOFTWARE OR ANY OTHER KIND OF
-- ASIC OR PROGRAMMABLE LOGIC DESIGN WITHOUT THE EXPLICIT PERMISSION OF THE
-- COPYRIGHT HOLDER!
--------------------------------------------------------------------------------
-- disclaimer of warranty:
--------------------------------------------------------------------------------
-- THIS SOURCECODE IS DISTRIBUTED IN THE HOPE THAT IT WILL BE USEFUL, BUT
-- THERE IS NO WARRANTY OR SUPPORT FOR THIS SOURCECODE. THE COPYRIGHT HOLDER 
-- PROVIDES THIS SOURCECODE "AS IS" WITHOUT WARRANTY OF ANY KIND,
-- EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
-- WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
-- THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THIS SOURCECODE IS WITH
-- YOU. SHOULD THIS SOURCECODE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL 
-- NECESSARY SERVICING, REPAIR OR CORRECTION.
--
-- IN NO EVENT WILL THE COPYRIGHT HOLDER BE LIABLE TO YOU FOR DAMAGES, INCLUDING
-- ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
-- USE OR INABILITY TO USE THIS SOURCECODE (INCLUDING BUT NOT LIMITED TO LOSS OF
-- DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD 
-- PARTIES OR A FAILURE OF THIS SOURCECODE TO OPERATE WITH ANY OTHER
-- SOFTWARE-PROGRAMS, HARDWARE-CIRCUITS OR ANY OTHER KIND OF ASIC OR
-- PROGRAMMABLE LOGIC DESIGN), EVEN IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF 
-- THE POSSIBILITY OF SUCH DAMAGES.
--------------------------------------------------------------------------------


library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;


entity fx2_slfifo_ctrl is
  generic
  (
    use_xil_dram : boolean := false;
    nr_of_dbgports : positive := 1
  );
  port
  (
    rst_i : in std_logic;
    
    -- ports connected to cypress fx2 device
    fx2_ifclk_i : in std_logic;
    fx2_flag_n_i : in std_logic_vector(2 downto 0);
    fx2_sloe_n_o : out std_logic;
    fx2_slrd_n_o : out std_logic;
    fx2_slwr_n_o : out std_logic;
    fx2_pktend_n_o : out std_logic;
    fx2_fifoadr_o : out std_logic_vector(1 downto 0);
    fx2_fd_io : inout std_logic_vector(15 downto 0);

    -- ports connected to user application
    app_fifo_wr_i : in std_logic;
    app_fifo_wr_data_i : in std_logic_vector(15 downto 0);
    app_fifo_wr_full_o : out std_logic;
    app_fifo_wr_count_o : out std_logic_vector(3 downto 0);
    app_usb_p2h_pktcount_o : out std_logic_vector(7 downto 0);

    app_fifo_rd_i : in std_logic;
    app_fifo_rd_data_o : out std_logic_vector(15 downto 0);
    app_fifo_rd_empty_o : out std_logic;
    app_fifo_rd_count_o : out std_logic_vector(3 downto 0);
    app_usb_h2p_pktcount_o : out std_logic_vector(7 downto 0);
        
    debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
  );
end fx2_slfifo_ctrl;


architecture RTL of fx2_slfifo_ctrl is

  signal rst : std_logic := '1';
  
  signal fx2_ifclk : std_logic := '0';
  signal fx2_flag_n : std_logic_vector(2 downto 0) := (others => '1');
  signal fx2_sloe_n : std_logic := '1';
  signal fx2_slrd_n : std_logic := '1';
  signal fx2_slwr_n : std_logic := '1';
  signal fx2_pktend_n : std_logic := '1';
  signal fx2_fifoadr : std_logic_vector(1 downto 0) := (others => '0');
  constant INPUT : std_logic := '0';
  constant OUTPUT : std_logic := not INPUT;
  type rec_fx2_fd is record
    oe : std_logic;
    output : std_logic_vector(15 downto 0);
    input : std_logic_vector(15 downto 0);
  end record;
  constant fx2_fd_default : rec_fx2_fd :=
  (
    oe => INPUT,
    output => (others => '0'),
    input => (others => '0')
  );
  signal fx2_fd : rec_fx2_fd := fx2_fd_default;

  signal app_fifo_wr : std_logic := '0';
  signal app_fifo_wr_data : std_logic_vector(15 downto 0) := (others => '0');
  signal app_fifo_wr_full : std_logic := '0';
  signal app_fifo_wr_count : std_logic_vector(3 downto 0) := (others => '0');
  signal app_usb_p2h_pktcount : unsigned(7 downto 0) := (others => '0');

  signal app_fifo_rd : std_logic := '0';
  signal app_fifo_rd_data : std_logic_vector(15 downto 0) := (others => '0');
  signal app_fifo_rd_empty : std_logic := '0';
  signal app_fifo_rd_count : std_logic_vector(3 downto 0) := (others => '0');
  signal app_usb_h2p_pktcount : unsigned(7 downto 0) := (others => '0');
  
  component sync_fifo16 is
    generic
    (
      data_width : positive := 16;
      use_xil_dram : boolean := false;
      nr_of_dbgports : positive := 1
    );
    port
    (
      rst : in std_logic;
      clk : in std_logic;
  
      wr_i : in std_logic;
      wr_data_i : in std_logic_vector((data_width-1) downto 0);
      wr_full_o : out std_logic;
  
      rd_i : in std_logic;
      rd_data_o : out std_logic_vector((data_width-1) downto 0);
      rd_empty_o : out std_logic;
  
      count_o : out std_logic_vector(3 downto 0);
  
      debug : inout std_logic_vector((nr_of_dbgports-1) downto 0)
    );
  end component;

  signal clk : std_logic := '0';

  type arr_fifo_data is array(natural range <>) of std_logic_vector(15 downto 0);
  signal fifo_wr : std_logic_vector(1 downto 0) := (others => '0');
  signal fifo_wr_data : arr_fifo_data(1 downto 0) := (others => (others => '0'));
  signal fifo_rd : std_logic_vector(1 downto 0) := (others => '0');
  signal fifo_rd_data : arr_fifo_data(1 downto 0) := (others => (others => '0'));

  signal fifo_wr_full : std_logic_vector(1 downto 0) := (others => '1');
  signal fifo_rd_empty : std_logic_vector(1 downto 0) := (others => '1');
  type arr_fifo_count is array(natural range <>) of std_logic_vector(3 downto 0);
  signal fifo_count : arr_fifo_count(1 downto 0) := (others => (others => '0'));
  
  constant H2P : natural := 0;
  constant P2H : natural := 1;
  
  constant empty : natural := 0;
  constant almost_empty : natural := 1;
  constant almost_full : natural := 2;

  type t_states is 
  (
    S_H2P,
    S_H2P_RECOVERY0,
    S_H2P_RECOVERY1,
    S_H2P_RECOVERY2,
    S_P2H,
    S_P2H_RECOVERY0,
    S_P2H_RECOVERY1,
    S_P2H_RECOVERY2
  );
  signal state : t_states := S_P2H_RECOVERY1;
  
  signal fx2_slrd : std_logic_vector(1 downto 0) := (others => '0');
  signal fx2_slwr : std_logic := '0';
  
  signal h2p_possible : std_logic := '0';
  signal p2h_possible : std_logic := '0';

begin

  -- connect entity ports to internal signals
  -------------------------------------------
  
  rst <= rst_i;
  
  fx2_ifclk <= fx2_ifclk_i;
  fx2_flag_n <= fx2_flag_n_i;
  fx2_sloe_n_o <= fx2_sloe_n;
  fx2_slrd_n_o <= fx2_slrd_n;
  fx2_slwr_n_o <= fx2_slwr_n;
  fx2_pktend_n_o <= fx2_pktend_n;
  fx2_fifoadr_o <= fx2_fifoadr;
  process(fx2_fd)
  begin
    if fx2_fd.oe = OUTPUT then
      fx2_fd_io <= fx2_fd.output;
    else
      fx2_fd_io <= (others => 'Z');
    end if;
  end process;
  fx2_fd.input <= fx2_fd_io;
    
  app_fifo_wr <= app_fifo_wr_i;
  app_fifo_wr_data <= app_fifo_wr_data_i;
  app_fifo_wr_full_o <= app_fifo_wr_full;
  app_fifo_wr_count_o <= app_fifo_wr_count;
  app_usb_p2h_pktcount_o <= std_logic_vector(app_usb_p2h_pktcount);

  app_fifo_rd <= app_fifo_rd_i;
  app_fifo_rd_data_o <= app_fifo_rd_data;
  app_fifo_rd_empty_o <= app_fifo_rd_empty;
  app_fifo_rd_count_o <= app_fifo_rd_count;
  app_usb_h2p_pktcount_o <= std_logic_vector(app_usb_h2p_pktcount);

  debug <=
  (
    others => 'Z'
  );
  
  -- instantiate small fpga buffers
  ---------------------------------
  
  fifos:
  for i in 0 to 1 generate
    inst_sync_fifo16 : sync_fifo16
    generic map
    (
      data_width => 16,
      use_xil_dram => use_xil_dram,
      nr_of_dbgports => 1
    )
    port map
    (
      rst => rst,
      clk => clk,
  
      wr_i => fifo_wr(i),
      wr_data_i => fifo_wr_data(i),
      wr_full_o => fifo_wr_full(i),
  
      rd_i => fifo_rd(i),
      rd_data_o => fifo_rd_data(i),
      rd_empty_o => fifo_rd_empty(i),
  
      count_o => fifo_count(i),
  
      debug => open
    );
  end generate;
  
  -- buffer connections to user application
  fifo_rd(H2P) <= app_fifo_rd;
  app_fifo_rd_data <= fifo_rd_data(H2P);
  app_fifo_rd_empty <= fifo_rd_empty(H2P);
  app_fifo_rd_count <= fifo_count(H2P);
    
  fifo_wr(P2H) <= app_fifo_wr;
  fifo_wr_data(P2H) <= app_fifo_wr_data;
  app_fifo_wr_full <= fifo_wr_full(P2H);
  app_fifo_wr_count <= fifo_count(P2H);

  -- architecture implementation
  ------------------------------
  
  -- single clock domain, fully synchronous to interface clock
  clk <= fx2_ifclk;
  
  -- connection external/internal buffers (combinatorial for peripheral->host path, 
  -- because fifo outputs are already registered)
  fx2_fd.output <= fifo_rd_data(P2H);
  fx2_fifoadr(0) <= '0'; -- fixed, because only two fx2 endpoints used (ep2 out(h2p) a[1:0]=b"00", ep6 in(p2h) a[1:0]=b"10")
  fx2_pktend_n <= '1'; -- no short packet transfer, the usb packet size is fixed at 512 bytes, !must! be tied to '1'
  
  fifo_rd(P2H) <= fx2_slwr;
  p2h_possible <= fx2_flag_n(almost_full) and (not fifo_rd_empty(P2H));

  process
  (
    rst, clk, 
    h2p_possible, p2h_possible, fx2_flag_n, fifo_count, state, fx2_slrd
  )
  begin

    -- clocked process part
    -----------------------
    
    if rst = '1' then
      app_usb_p2h_pktcount <= (others => '0');
      app_usb_h2p_pktcount <= (others => '0');
      fx2_fd.oe <= INPUT;
      fx2_fifoadr(1) <= '0';
      fx2_sloe_n <= '1';
      fx2_slrd_n <= '1';
      fx2_slwr_n <= '1';
      state <= S_P2H_RECOVERY1;
      fifo_wr(H2P) <= '0';
      fx2_slrd(1) <= '0';
    elsif clk'event and clk = '1' then
      -- connection external/internal buffers (registered for host->peripheral path)
      fifo_wr_data(H2P) <= fx2_fd.input;
      -- all usb transfers are aligned to 256 16-bit word packets
      -- counters for the word index to determine, if an usb packet is completely 
      -- received or transmitted
      if app_fifo_wr = '1' then
        app_usb_p2h_pktcount <= app_usb_p2h_pktcount + 1;
      end if;
      if app_fifo_rd = '1' then
        app_usb_h2p_pktcount <= app_usb_h2p_pktcount + 1;
      end if;
      -- synthesize flip-flops for external slave fifo control signals to improve i/o timing
      fx2_slrd_n <= not fx2_slrd(0);
      fx2_slwr_n <= not fx2_slwr;
      fx2_slrd(1) <= fx2_slrd(0);
      fifo_wr(H2P) <= fx2_slrd(1);
      -- fsm copying fx2 slave fifo endpoint buffers to
      -- fpga internal fifo buffers
      -- 3 recovery states between a data direction change ensure a very relaxed
      -- timing for fx2/fpga output enables and fifo address
      -- data direction changes, if there is no data transfer in current direction,
      -- but data transfer in the other direction would be possible
      case state is
        when S_H2P =>
          if ((not h2p_possible) and p2h_possible) = '1' then
            state <= S_H2P_RECOVERY0;
          end if;
        when S_H2P_RECOVERY0 =>
          fx2_sloe_n <= '1';
          fx2_fifoadr(1) <= '1';
          state <= S_H2P_RECOVERY1;
        when S_H2P_RECOVERY1 =>
          fx2_fd.oe <= OUTPUT;
          state <= S_H2P_RECOVERY2;
        when S_H2P_RECOVERY2 =>
          state <= S_P2H;
        when S_P2H =>
          if ((not p2h_possible) and h2p_possible) = '1' then
            state <= S_P2H_RECOVERY0;
          end if;
        when S_P2H_RECOVERY0 =>
          fx2_fd.oe <= INPUT;
          fx2_fifoadr(1) <= '0';
          state <= S_P2H_RECOVERY1;
        when S_P2H_RECOVERY1 =>
          fx2_sloe_n <= '0';
          state <= S_P2H_RECOVERY2;
        when S_P2H_RECOVERY2 =>
          state <= S_H2P;
        when others => null;
      end case;
    end if;

    -- combinatorial process part
    -----------------------------
    
    h2p_possible <= '0';
    -- check almost empty flag in data transfers, because i/o ff of slave read strobe
    -- causes one clk cycle delay, another clk cycle delay because of fifo_wr_data(H2P)
    -- register
    -- => (fifo count < 13) used instead of almost full or full flag
    if fx2_slrd(1) = '1' then
      if (fx2_flag_n(almost_empty) = '1') and (unsigned(fifo_count(H2P)) < 13) then
        h2p_possible <= '1';
      end if;
    -- be sure to take last value out of fx2 fifo, if no data transfer is in progress
    -- in current clock cycle
    else
      if (fx2_flag_n(empty) = '1') and (unsigned(fifo_count(H2P)) < 13) then
        h2p_possible <= '1';
      end if;
    end if;

    -- apply external fifo read- and write-strobe signals in appropriate fsm state
    fx2_slrd(0) <= '0';
    fx2_slwr <= '0';
    case state is
      when S_H2P =>
        fx2_slrd(0) <= h2p_possible;
      when S_P2H =>
        fx2_slwr <= p2h_possible;
      when others => null;
    end case;
    
  end process;

end RTL;